from contextlib import contextmanager
from fractions import Fraction
from pathlib import Path
import sys
from typing import Annotated, Literal, TextIO
from cyclopts import App, Parameter
import tomli

from zucchini.exceptions import BrokenAutograderError
from zucchini.exporters import EXPORTERS, ExporterKey
from zucchini.utils import glob_get_files

from .assignment import Assignment, AssignmentConfig, AssignmentMetadata
from .grades import AssignmentGrade, ComponentGrade
from .gradescope import GradescopeMetadata

from .submission import Submission

app = App(
    help="A fun autograder management system for the whole family.",
    help_format="md",
    help_formatter="plain",
    help_on_error=True,
)

ASSIGNMENT_FILES_DIR = "grading-files"
ASSIGNMENT_CONFIG_TOML = "zucchini.toml"
ASSIGNMENT_CONFIG_JSON = "zucchini.json"

def _get_assignment_cfg(grading_dir: Path) -> AssignmentConfig:
    # Try reading from TOML
    try:
        with open(grading_dir / ASSIGNMENT_CONFIG_TOML, "rb") as f:
            data = tomli.load(f) 
            return AssignmentConfig.model_validate(data)
    except FileNotFoundError:
        pass

    # Try reading from JSON
    try:
        with open(grading_dir / ASSIGNMENT_CONFIG_JSON, "rb") as f:
            return AssignmentConfig.model_validate_json(f.read())
    except FileNotFoundError:
        pass

    raise FileNotFoundError(f"Missing configuration file {ASSIGNMENT_CONFIG_TOML} in {grading_dir}")
def _get_assignment(submission_dir: Path, grading_dir: Path, gs_metadata_fp: Path) -> tuple[Assignment, Submission]:
    # Get metadata:
    submission_created_at = None
    metadata = AssignmentMetadata()

    # Load metadata from Gradescope submission_metadata.json file:
    try:
        with open(gs_metadata_fp, "rb") as f:
            gs_metadata = GradescopeMetadata.model_validate_json(f.read())
            
            metadata = gs_metadata.as_metadata()
            submission_created_at = gs_metadata.created_at
    except FileNotFoundError:
        pass

    # Get config:
    config = _get_assignment_cfg(grading_dir)
    return (
        Assignment(config, grading_dir / ASSIGNMENT_FILES_DIR, metadata),
        Submission(submission_dir, submission_created_at)
    )

@contextmanager
def _io_manage(path: Path | None, fallback: TextIO, rw: Literal["r", "w"]):
    resource = open(path, rw) if path is not None else fallback
    try:
        yield resource
    finally:
        if resource is not fallback:
            resource.close()

@app.command()
def verify_config(autograder_path: Path = Path(".")):
    """
    Verifies the Zucchini config in the specified directory is valid.

    Parameters
    ----------
    autograder_path: Path, optional
        The path the config is located.

        This path should contain a Zucchini configuration file (zucchini.toml)
        and a "grading-files/" folder which contains the autograder test cases.
    """
    cfg = _get_assignment_cfg(autograder_path)

    # Checks configs are correct:
    for c in cfg.components:
        glob_get_files(c.grading_files, autograder_path)

    print(cfg.model_dump_json())

@app.command()
def grade(
    submission_path: Path = Path("/autograder/submission"),
    metadata_path: Path = Path("/autograder/submission_metadata.json"),
    *,
    autograder_path: Path = Path("."),
    output_: Annotated[Path | None, Parameter(alias="-o")] = None,
):
    """
    Grades a submission, located at the specified submission path.

    Parameters
    ----------
    submission_path : Path, optional
        The path of the submission files.
    
    metadata_path : Path, optional
        The path of the submission metadata.

        This is produced by Gradescope and is typically called "submission_metadata.json".

    autograder_path : Path, optional
        The path the autograder is located.

        This path should contain a Zucchini configuration file (zucchini.toml)
        and a "grading-files/" folder which contains the autograder test cases.
    
    output_ : Path, optional
        The output path to use. If unspecified, this defaults to sys.stdout.
    """
    try:
        assignment, submission = _get_assignment(submission_path, autograder_path, metadata_path)
        grade = assignment.grade(submission)
    except Exception as e:
        ze = BrokenAutograderError(str(e))
        grade = AssignmentGrade(
            name="Assignment",
            raw_score=Fraction(0),
            final_score=Fraction(0),
            max_points=Fraction(0),
            components=[ComponentGrade(
                norm_weight=Fraction(1),
                description="Unexpected error",
                error=ze
            )],
            penalties=[]
        )

    with _io_manage(output_, sys.stdout, "w") as out:
        print(grade.model_dump_json(), file=out)

@app.command()
def export(
    format_: ExporterKey,
    *,
    input_: Annotated[Path | None, Parameter(alias="-i")] = None,
    output_: Annotated[Path | None, Parameter(alias="-o")] = None
):
    """
    Converts a Zucchini component grades result (which would be obtained from "zucc grade")
    and converts it to some supported export format.
    
    Parameters
    ----------
    format_ : ExporterKey
        The format to use.

        - The "local" format is useful for terminals, displaying results in a printable format.
        - The "gradescope" format is useful for Gradescope, displaying results in the JSON format Gradescope requires.

    input_ : Path, optional
        The input path to use. If unspecified, this defaults to sys.stdin.

    output_ : Path, optional
        The output path to use. If unspecified, this defaults to sys.stdout.
    """
    with _io_manage(input_, sys.stdin, "r") as inp:
        grade = AssignmentGrade.model_validate_json(inp.read())

    Exporter = EXPORTERS.get(format_)
    if Exporter is None:
        raise ValueError(f"Invalid format {format_!r}")
    exporter = Exporter()
    
    with _io_manage(output_, sys.stdout, "w") as out:
        exporter.export(grade, out)
        print(file=out)

if __name__ == "__main__":
    app()


import json
from fractions import Fraction
from pathlib import Path
import subprocess
from typing import Literal
from typing_extensions import override

from ..exceptions import BrokenSubmissionError
from ..utils import run_command
from ..grades import PartGrade
from . import GraderInterface, Part

"""
Grade a homework using a CircuitSim grader
"""

class CircuitSimTest(Part):
    test: str
    """
    Name of test.
    This corresponds to the method name of the JUnit test.
    """
    
    @override
    def description(self):
        return self.test

    def grade(self, result):
        if result is None:
            msg = ('Results for test not found. Check if there were any'
                   ' internal errors reported. If not, report this as an'
                   ' autograder error to your instructors.')
            return PartGrade(score=Fraction(0), log=msg)

        log = '\n'.join('{}: {}'.format(
            failure['displayName'],
            failure.get('message', '(no details, sorry)'))
            for failure in result['partialFailures'])
        partialFailures = len(result['partialFailures'])
        if partialFailures < result['failed']:
            log += '\n[omitted {} more failures]' \
                   .format(result['failed'] - partialFailures)

        score = Fraction(result['total'] - result['failed'], result['total'])
        return PartGrade(score=score, log=log)

class CircuitSimGrader(GraderInterface[CircuitSimTest]):
    """
    Run a CircuitSim grader based on
    <https://github.com/ausbin/circuitsim-grader-template> and collect
    the results.
    """
    kind: Literal["CircuitSimGrader"]

    grader_jar: Path
    """
    Filename for the autograder JAR this autograder should run.
    """

    test_class: str
    """
    Name of class where tests are located.
    """

    timeout: float = 30
    """
    Timeout (in seconds) before the autograder aborts operation.
    """

    @override
    def list_prerequisites(self):
        # CircuitSim needs JavaFX
        return ['openjdk-8-jre', 'openjfx']

    @override
    def needs_display(self):
        # CircuitSim needs JavaFX which needs a display server
        return True

    @override
    def Part(self, _cd):
        return CircuitSimTest
    
    @override
    def grade(self, submission, path, parts):
        cmdline = ['java', '-jar', self.grader_jar, '--zucchini',
                   self.test_class]
        
        # Do not mix stderr into stdout because sometimes our friend
        # Roi printStackTrace()s or System.err.println()s, and that
        # will mess up JSON parsing
        cmd_result = run_command(cmdline, cwd=path, timeout=self.timeout, stderr=subprocess.PIPE)
        cmd_result.check_returncode()
        results = json.loads(cmd_result.stdout)

        if 'error' in results:
            raise BrokenSubmissionError(results['error'])

        method_results = {result['methodName']:
                          result for result in results['tests']}
        return [part.grade(method_results.get(part.test)) for part in parts]



from abc import ABC, abstractmethod
from fractions import Fraction
from pathlib import Path
from typing import Generic, TypeVar

from zucchini.grades import PartGrade
from zucchini.submission import Submission
from zucchini.utils import KebabModel, ShlexCommand

class Part(KebabModel, ABC):
    """
    Represents a 'part' of grading which has its own weight and its own
    score: one prompt question, one unit test, etc.
    """

    weight: Fraction
    """Weight of part"""

    partial_credit: bool = True
    """Whether partial credit should be acceptable"""

    @abstractmethod
    def description(self) -> str:
        """
        Return a human-friendly description for this part. Used in grade
        breakdowns and logs.
        """
        pass

P = TypeVar("P", bound=Part)
class GraderInterface(KebabModel, ABC, Generic[P]):
    def list_prerequisites(self) -> list[str]:
        """
        This function should return a list of Ubuntu 16.04 packages
        required to run this grader.
        """
        return []

    def list_extra_setup_commands(self) -> list[ShlexCommand]:
        """
        This function should return a list of extra one-time commands to
        run at Docker image creation time. This is Ubuntu.
        """
        return []

    def is_interactive(self) -> bool:
        """
        Return True if and only if this grader will produce command-line
        prompts.
        """
        return False

    def needs_display(self) -> bool:
        """
        Return True if and only if this grader expects a graphical
        environment, like $DISPLAY on GNU/Linux. Does not necessarily
        imply is_interactive() is True since some graders are
        noninteractive but still connect to a display server, like
        CircuitSim graders (since CircuitSim needs JavaFX).
        """
        return False
    
    @abstractmethod
    def Part(self, config_dict) -> type[P] | P:
        """
        The Part type of the grader.

        You may either return the class itself (e.g., `return Part`)
        or if you want to modify the default initialization behavior,
        then return a new instance of the Part (using the dictionary above).
        """

    def create_part(self, config_dict) -> P:
        """
        Creates a Part instance from a given configuration dictionary.

        This should not be overridden.
        If you wish to override default Part initialization behavior,
        overwrite `GraderInterface.Part` instead.
        """
        part = self.Part(config_dict)
        if isinstance(part, type):
            return part.model_validate(config_dict)
        return part
        
    @abstractmethod
    def grade(self, submission: Submission, path: Path, parts: list[P]) -> list[PartGrade]:
        """
        Grades a submission.

        Parameters
        ----------
        submission : Submission
            The submission object to grade.
        path : Path
            The path for the root submission directory.
            This is also the current directory, where commands are executed from.
        parts : list[P]
            A list of parts (from the config file).

            This will be the same type defined in `cls.Part()`
            and in the type argument for this type.

        Returns
        -------
        list[PartGrade]
            The grades for each individual part.
            There must be one for each part specified in the `parts` argument.
        """
        pass



import datetime as dt
import enum
import re
from fractions import Fraction
from typing import Annotated, Literal
from typing_extensions import override

from pydantic import BeforeValidator

from zucchini.utils import KebabModel

from . import PenalizerInterface, InvalidPenalizerConfigError

"""Penalize a late submission."""


class LatePenaltyType(enum.Enum):
    PERCENT = '%'
    POINTS = 'pts'
    MAX_POINTS = 'max_pts'

UNITS_REGEX = re.compile(r'^(?P<mag>[0-9/.]+)\s*(?P<unit>[a-z_-]*)$')
_UnitInput = str | int | float | Fraction

def _split_units(amount_str: _UnitInput) -> tuple[Fraction, str | None]:
    # Handle using the print version, instead of the literal float value
    if isinstance(amount_str, float):
        return Fraction(str(amount_str)), None
    # Treat as no unit
    if isinstance(amount_str, int | Fraction):
        return Fraction(amount_str), None
    
    match = UNITS_REGEX.match(amount_str.lower())

    if match is None:
        raise InvalidPenalizerConfigError(f"unknown units format {amount_str!r}")

    try:
        mag = Fraction(match.group("mag"))
    except ValueError:
        raise InvalidPenalizerConfigError(f"cannot parse magnitude")
    return mag, match.group('unit') or None

def _parse_timedelta(time_str: _UnitInput):
    # If failure to split, we just let data pass
    # to allow for other timedelta formats.
    try:
        mag, unit = _split_units(time_str)
    except InvalidPenalizerConfigError:
        return time_str
    
    unit = unit or "s"
    UNITS = {
        's': dt.timedelta(seconds=1),
        'm': dt.timedelta(minutes=1),
        'h': dt.timedelta(hours=1),
        'd': dt.timedelta(days=1)
    }
    if unit not in UNITS:
        raise InvalidPenalizerConfigError(f"unknown time unit {unit!r}. try one of {', '.join(UNITS)}.")
    
    return mag.numerator * UNITS[unit] / mag.denominator

def _parse_penalty(penalty_str: _UnitInput):
    mag, unit = _split_units(penalty_str)

    if unit in ('pt', 'pts'):
        return (mag / 100, LatePenaltyType.POINTS)
    elif unit in ('maxpts', 'max-pts', 'max_pts', 'maxpt', 'max-pt', 'max_pt'):
        return (mag / 100, LatePenaltyType.MAX_POINTS)
    elif unit is None:
        return (mag, LatePenaltyType.PERCENT)
    else:
        raise InvalidPenalizerConfigError(f"unknown penalty unit {unit!r}. try a fraction optionally followed by 'pts' or 'max-pts'.")

class LatePenalty(KebabModel):
    """
    A single penalty rule. 
    
    This applies a specified penalty only when the specified amount of time has passed.
    """

    after: Annotated[dt.timedelta, BeforeValidator(_parse_timedelta)]
    penalty: Annotated[tuple[Fraction, LatePenaltyType], BeforeValidator(_parse_penalty)]

    def is_late(self, duration_late: dt.timedelta | None):
        """Whether the submission is late under this penalty."""
        return duration_late is not None and duration_late > self.after

    def adjust_grade(self, grade: Fraction):
        """Adjust the grade as though the submission is late."""
        penalty, penalty_type = self.penalty
        match penalty_type:
            case LatePenaltyType.PERCENT:
                return grade * (1 - penalty)
            case LatePenaltyType.POINTS:
                return max(Fraction(0), grade - penalty)
            case LatePenaltyType.MAX_POINTS:
                return min(grade, penalty)
            case ty:
                raise ValueError(f"unknown penalty type {ty!r}. code bug?")

class LatePenalizer(PenalizerInterface):
    """
    Penalize students for late submissions.

    Configure it like this in the assignment config file. In this
    example, after 8 hours late you get 75 points off your grade (a 85
    would go to a 10).

    penalties:
    - name: LATE
      backend: LatePenalizer
      backend-options:
        penalties:
        - after: 1h
          penalty: 25pts
        - after: 8h
          penalty: 50pts

    That is, penalties are applied in order, and they do not stop when
    there is a match. They don't necessarily have to be in increasing
    order of `after' values.

    The penalty unit (pts above) can be any of the following, where N is the
    number provided (e.g., N pts):

     * pt, pts: Subtract N/100, a fixed number of points, from the student's
                grade. Do not allow their grade to go below 0.

     * maxpts, max_pts, maxpt, max_pt: Set the maximum grade a student can
                                       achieve to N/100.

     * (none): Multiply the student's grade by (1-N). Here, N is expected to be
               a fraction already, e.g., 1/4.
    """

    kind: Literal["LatePenalizer"]
    penalties: list[LatePenalty]
    
    apply_one: bool = False
    """
    Whether to apply only the latest applicable penalty.

    If True, this only applies the last penalty (chronologically) which is marked late.
    If False, this applies all penalties marked late.

    This field is particularly useful when thinking about penalties in terms of intervals.
    If penalties are ordered chronologically, 
    then a given penalty is applied if the submission time is between the current and next penalty.
    """

    @override
    def adjust_grade(self, grade: Fraction, submission, metadata):
        # Determine how late submission was:
        duration_late = None
        if submission.submit_date is not None and metadata.due_date is not None:
            duration_late = submission.submit_date - metadata.due_date
            
        # Apply penalties accordingly:
        penalties = sorted(self.penalties, key=lambda p: p.after)
        new_grade = grade
        for penalty in penalties:
            if penalty.is_late(duration_late):
                # Undo any penalty except last:
                if self.apply_one:
                    new_grade = grade
                new_grade = penalty.adjust_grade(new_grade)

        return new_grade



import datetime as dt

from fractions import Fraction
from typing import Annotated
from pydantic import BaseModel, Field

from zucchini.assignment import AssignmentMetadata


class GradescopeUserAssignment(BaseModel):
    release_date: dt.datetime
    """Release date of the assignment for the user."""
    due_date: dt.datetime
    """Due date of the assignment for the user."""
    late_due_date: dt.datetime | None = None
    """Late due date of the assignment for the user."""

class GradescopeUser(BaseModel):
    user_id: Annotated[int, Field(alias="id")]
    """
    ID of author.

    This ID is used exclusively by Gradescope.
    """

    email: str
    """Email of author."""
    name: str
    """Name of author."""
    sid: int
    """
    Student ID.

    This ID is used throughout multiple systems,
    including those outside Gradescope.
    """
    assignment: GradescopeUserAssignment
    """Assignment metadata for the user."""
    sections: list[str]
    """Sections the author are a part of."""

class GradescopeCurrentAssignmentOutlineItem(BaseModel):
    outline_id: Annotated[int, Field(alias="id")]
    weight: Fraction
    # There are other fields, but I don't think it's worth serializing these

class GradescopeCurrentAssignment(BaseModel):
    assignment_id: Annotated[int, Field(alias="id")]
    """
    ID of the assignment.
    """

    title: str
    """Title of the assignment."""

    release_date: dt.datetime
    """Default release date of the assignment."""
    due_date: dt.datetime
    """Default due date of the assignment."""
    late_due_date: dt.datetime | None = None
    """Default late due date of the assignment."""
    total_points: Fraction
    """Total points for the assignment."""
    course_id: int
    """Course ID (used exclusively by Gradescope)."""
    group_submission: bool
    """Whether this assignment is a group submission."""
    group_size: int | None
    """Group size of assignment (None if `group_submission` is False)"""
    outline: list[GradescopeCurrentAssignmentOutlineItem]
    """Outline of tasks in assignment."""

class GradescopePreviousSubmission(BaseModel):
    submission_id: Annotated[int, Field(alias="id")]
    """ID of submission."""

    submission_time: dt.datetime
    """Datetime when the submission was created."""

    score: Fraction
    """Score received for submission."""

    autograder_error: bool
    """Whether the autograder ran correctly."""

    # results: list[...]
    # Don't want to implement this

class GradescopeMetadata(BaseModel):
    submission_id: Annotated[int, Field(alias="id")]
    """ID of submission."""

    users: list[GradescopeUser]
    """List of authors of the submission."""
    
    created_at: dt.datetime
    """Datetime when the submission was created."""

    # assignment_id: str | None
    # No idea what this is

    assignment: GradescopeCurrentAssignment
    """Data for current assignment"""

    previous_submissions: list[GradescopePreviousSubmission]
    """Previous submissions."""

    def due_date(self):
        """Due date for the assignment (based on the users' due dates)."""
        return max(u.assignment.due_date for u in self.users)
    
    def as_metadata(self):
        return AssignmentMetadata(
            total_points=self.assignment.outline[0].weight,
            due_date=self.due_date(),
        )


import json
import re
import tempfile
from typing import Literal
from typing_extensions import override

from ..utils import ShlexCommand, run_command
from ..grades import PartGrade, Fraction
from . import Part, ThreadedGrader

class CriterionTest(Part):
    name: str
    """
    Name of test.

    This is a descriptor which will be displayed on Gradescope.
    """

    suite: str
    """
    The suite identifier.
    """

    test: str = "*"
    """
    The test identifier.

    This is the identifier used to distinguish test cases within a suite.
    """

    valgrind_deduction: Fraction = Fraction(1, 2)

    @override
    def description(self):
        return self.name
    
    def grade(self, path, grader):
        result_file = tempfile.NamedTemporaryFile(prefix='zlog-', suffix='.json', dir=path, delete=True)
        result_file.close() # Close file to allow subprocess to write

        flags = [
            "--timeout", str(grader.single_timeout),
            f"--filter={self.suite}/{self.test}",
            f"--json={result_file.name}"
        ]

        command = ['./tests', *flags]
        valgrind_cmd = None
        if grader.valgrind_cmd is not None:
            valgrind_cmd = grader.valgrind_cmd + flags

        # Run tests without Valgrind:
        cmd_result = run_command(command, cwd=path, timeout=grader.total_timeout)
        out = cmd_result.stdout

        with open(result_file.name, "r") as f:
            try:
                report_data = json.load(f)
            except json.JSONDecodeError:
                return PartGrade(score=0, log="Cannot parse result JSON")
        passing = report_data["passed"]
        failing = report_data["failed"]
        total = passing + failing
        
        if total == 0:
            return PartGrade(score=0, log="No test cases were found")

        if total == passing:
            # Standard autograder passed, so check Valgrind:
            # Check Valgrind test:
            if valgrind_cmd is not None:
                cmd_result = run_command(valgrind_cmd, cwd=path, timeout=grader.total_timeout)
                out = cmd_result.stdout

                if re.search(r"^==\d+==.*$", out, re.MULTILINE): # Valgrind failed
                    return PartGrade(score=1 - self.valgrind_deduction, log=out)
            return PartGrade(score=1, log="")
        
        log_line_pattern = r"^\s*\[(?:----|FAIL)\].*"
        log = "\n".join(re.findall(log_line_pattern, out, re.MULTILINE))
        deduct_factor = 1 # factor added to make Valgrind failures fairer
        if valgrind_cmd is not None:
            log += "\n\nValgrind skipped due to failed test."
            deduct_factor *= 1 - self.valgrind_deduction

        return PartGrade(score=(passing / total) * deduct_factor, log=log)

class CriterionGrader(ThreadedGrader[CriterionTest]):
    kind: Literal["CriterionGrader"]

    single_timeout: float = 3
    """Timeout for a single test."""

    total_timeout: float = 60
    """Timeout for the autograder overall."""

    valgrind_cmd: ShlexCommand | None = None
    """Command to use to run Valgrind."""
    
    @override
    def list_prerequisites(self):
        return []
    
    @override
    def Part(self, _cd):
        return CriterionTest
    
    @override
    def grade_part(self, part, path, submission):
        return part.grade(path, self)

    @override
    def grade(self, submission, path, parts):
        # Compile program:
        cmd_result = run_command(["make", "tests"], cwd=path, timeout=self.total_timeout)
        cmd_result.check_returncode({ 0, 1 })

        # Run each part:
        return super(CriterionGrader, self).grade(submission, path, parts)


import json
import tempfile
from fractions import Fraction
from typing import Literal
from typing_extensions import override

from ..exceptions import BrokenSubmissionError
from ..utils import run_command
from ..grades import PartGrade
from . import GraderInterface, Part

"""
Grade a homework using the classic Java bitwise operators grader
"""

class BitwiseJSONMethod(Part):
    """
    A method in the bitwise JSON autograder.
    """

    class_name: str | None = None
    """Name of the class where the method is located."""

    method: str
    """The name of the method."""

    @override
    def description(self):
        return f'{self.class_name}.{self.method}()'

    def grade(self, jsonResults):
        # If the method is missing from the results, they probably
        # didn't implement it.
        if self.method not in jsonResults:
            return PartGrade(score=Fraction(0), log='method not found',
                             deductions=['missing'])
        else:
            jsonResult = jsonResults[self.method]

        if jsonResult.get('errorMessage', None):
            return PartGrade(score=Fraction(0), log=jsonResult['errorMessage'])

        violations: list[str] = jsonResult.get('violations', [])

        # If they violated any rules, boom, zero
        if violations:
            score = Fraction(0)
        else:
            score = Fraction(jsonResult['testsPassed'],
                             jsonResult['testsTotal'])

        log = jsonResult.get('message', '')

        return PartGrade(score=score, deductions=violations, log=log)

class BitwiseJSONGrader(GraderInterface[BitwiseJSONMethod]):
    """
    Run the legendary hw2checker.jar and collect results.
    """
    
    kind: Literal["BitwiseJSONGrader"]
    
    grader_jar: str
    """
    Filename for the autograder JAR this autograder should run.
    """

    source_file: str
    """
    The source file we're autograding (this will be the user's file).
    """

    timeout: float = 10
    """
    Timeout (in seconds) before the autograder aborts operation.
    """

    @override
    def list_prerequisites(self):
        return ['openjdk-8-jre-headless']

    def class_name(self):
        """Return the class name of this file"""
        return self.source_file.rsplit('.', maxsplit=1)[0]
    
    @override
    def Part(self, config_dict):
        return BitwiseJSONMethod.model_validate({ "class_name": self.class_name(), **config_dict })
    
    @override
    def grade(self, submission, path, parts):
        result_file = tempfile.NamedTemporaryFile(prefix='zlog-', suffix='.json', dir=path, delete=True)
        result_file.close() # Close file to allow subprocess to write

        # Run command:
        cmdline = ['java', '-jar', self.grader_jar, '-z', self.source_file, result_file.name]
        cmd_result = run_command(cmdline, cwd=path, timeout=self.timeout)
        cmd_result.check_returncode()

        with open(result_file.name) as gradelog_file:
            gradelog = json.load(gradelog_file)

        if gradelog.get('errorMessage', None):
            raise BrokenSubmissionError(gradelog['errorMessage'])

        return [part.grade(gradelog['results']) for part in parts]



from fractions import Fraction
from typing import Any, Literal
from pydantic import BaseModel, TypeAdapter

from zucchini.exceptions import ZucchiniError
from zucchini.grades import AssignmentGrade, ComponentGrade

OutputVisibility = Literal["hidden", "after_due_date", "after_published", "visible"]
"""
Visibility of an assignment grade, test, or output.

<https://gradescope-autograders.readthedocs.io/en/latest/specs/#controlling-test-case-visibility>
"""
TestStatus = Literal["passed", "failed"]
"""
Status of a given test case.

<https://gradescope-autograders.readthedocs.io/en/latest/specs/#test-case-status>
"""
OutputFormat = Literal["text", "html", "simple_format", "md", "ansi"]
"""
Format of text.

<https://gradescope-autograders.readthedocs.io/en/latest/specs/#output-string-formatting>
"""

def _get_status(passed: bool) -> TestStatus:
    """
    Override how Gradescope determines status.
    
    If a student gets 0% of 0 points, it is interpreted as a pass
    (even if the test failed).

    To circumvent this, we instead use the grade's `passed` variables,
    which keep track of when the tests succeed
    (as opposed to keeping track of weighted score).
    """
    return "passed" if passed else "failed"

def _error_output(error: ZucchiniError):
    base = f"{error}\n{error.verbose or ''}".strip()
    if error.is_it_autograders_fault:
        ANSI_RED ="\x1b[31m"
        ANSI_RESET = "\x1b[0m"
        base += (
            f"\n\n{ANSI_RED}This appears to be an autograder bug. "
            f"Please report this as an autograder error to your instructors.{ANSI_RESET}"
        )
    return base

def _gs_float(f: Fraction | int):
    """
    Converts the value to a simplified decimal,
    to prevent long floats from displaying in Gradescope.
    """
    # FIXME(3.12): Fraction.__format__ begins supporting .02f syntax in Py3.12
    return f"{float(f):.02f}"

class GradescopeTestOutput(BaseModel):
    score: str | int | None = None
    max_score: str | int | None = None
    status: TestStatus | None = None
    name: str | None = None
    name_format: OutputFormat | None = None
    number: str | None = None
    output: str | None = None
    output_format: OutputFormat | None = None
    tags: list[str] | None = None
    visibility: OutputVisibility | None = None
    extra_data: dict[Any, Any] | None = None

class GradescopeOutput(BaseModel):
    """
    Gradescope autograder output, as specified in:
    <https://gradescope-autograders.readthedocs.io/en/latest/specs/#output-format>
    """

    score: str | int | None = None
    execution_time: int | None = None
    output: str | None = None
    output_format: OutputFormat | None = None
    test_output_format: OutputFormat | None = None
    test_name_format: OutputFormat | None = None
    visibility: OutputVisibility | None = None
    stdout_visibility: OutputVisibility | None = None
    extra_data: dict[Any, Any] | None = None
    tests: list[GradescopeTestOutput] | None = None

    @classmethod
    def from_grade(cls, grade: AssignmentGrade):
        tests: list[GradescopeTestOutput] = []
        
        # Penalties:
        for penalty in grade.penalties:
            if penalty.points_deducted != 0:
                score = _gs_float(-max(0, penalty.points_deducted) * grade.max_points)
                tests.append(GradescopeTestOutput(
                    name=penalty.name,
                    score=score,
                    max_score=0,
                    status="failed" if penalty.points_deducted > 0 else "passed",
                ))
        
        # Actual test results
        for component in grade.components:
            if component.error:
                # Broken error
                score = _gs_float(component.points_received() * grade.max_points)
                max_score = _gs_float(component.norm_weight * grade.max_points)

                tests.append(GradescopeTestOutput(
                    name=component.description,
                    score=score,
                    max_score=max_score,
                    status=_get_status(component.passed()),
                    output=_error_output(component.error),
                ))
            else:
                # Tests
                for part in (component.parts or []):
                    score = _gs_float(part.points_received() * component.norm_weight * grade.max_points)
                    max_score = _gs_float(part.norm_weight * component.norm_weight * grade.max_points)

                    deductions = ""
                    if part.inner.deductions:
                        deductions = f"Deductions: {', '.join(part.inner.deductions)}\n\n"
                    output = deductions + (part.inner.log or "")

                    tests.append(GradescopeTestOutput(
                        name=f"{component.description}: {part.description}",
                        score=score,
                        max_score=max_score,
                        status=_get_status(part.passed()),
                        output=output
                    ))

        extra_data = {
            "component_grades": TypeAdapter(list[ComponentGrade]).dump_json(grade.components)
        }
        return cls(
            score=_gs_float(grade.final_grade()),
            tests=tests,
            extra_data=extra_data,
            output_format="ansi",
            test_output_format="ansi",
        )


from abc import ABC, abstractmethod
from fractions import Fraction
from typing import TYPE_CHECKING

from zucchini.utils import KebabModel

if TYPE_CHECKING:
    from zucchini.assignment import AssignmentMetadata
from zucchini.submission import Submission

class PenalizerInterface(KebabModel, ABC):
    """Penalize a student outside component grades. Example: Late penalties"""
    
    @abstractmethod
    def adjust_grade(self, grade: Fraction, submission: Submission, metadata: "AssignmentMetadata") -> Fraction:
        """
        This function should take in a Submission object and a Fraction
        holding the calculated grade, returning the new grade.
        """
        pass



from typing import Union, get_args
from ..exceptions import InvalidGraderConfigError
from .grader_interface import GraderInterface, Part
from .threaded_grader import ThreadedGrader
from .bitwise_json_grader import BitwiseJSONGrader
from .circuitsim_grader import CircuitSimGrader
from .multi_command_grader import MultiCommandGrader
from .ensemble_grader import EnsembleGrader
from .criterion_grader import CriterionGrader

__all__ = ['InvalidGraderConfigError', 'GraderInterface', 'Part',
           'ThreadedGrader', 'BitwiseJSONGrader', 'CircuitSimGrader', 
           'MultiCommandGrader', 'EnsembleGrader', 'CriterionGrader']

SupportedGrader = Union[
    BitwiseJSONGrader,
    CircuitSimGrader,
    MultiCommandGrader,
	EnsembleGrader,
	CriterionGrader
]
AVAILABLE_GRADERS = {cls.__name__: cls for cls in get_args(SupportedGrader)}



from pathlib import Path
import tempfile
from typing import Literal
from typing_extensions import override

import xml.etree.ElementTree
from ..utils import run_command
from ..grades import PartGrade
from . import Part, GraderInterface


class EnsembleTest(Part):
    name: str
    """
    Name of test.

    This corresponds to the function name used by Pytest.
    """

    @override
    def description(self):
        return self.name

    def grade(self, result):
        if result is None:
            msg = ('Results for test not found. Check if there were any'
                   ' internal errors reported. If not, report this as an'
                   ' autograder error to your instructors.')
            return PartGrade(score=0, log=msg)

        failure = result.find('failure')

        if failure is not None:
            log = "\n".join([failure.get('message'), failure.text])
            return PartGrade(score=0, log=log)

        return PartGrade(score=1, log='')

class EnsembleGrader(GraderInterface[EnsembleTest]):
    kind: Literal["EnsembleGrader"]

    test_file: Path
    """Name of file to test on."""

    timeout: float = 30
    """Timeout for autograder."""

    @override
    def list_prerequisites(self):
        return []

    @override
    def Part(self, _cd):
        return EnsembleTest
    
    @override
    def grade(self, submission, path, parts):
        result_file = tempfile.NamedTemporaryFile(prefix='zlog-', suffix='.xml', dir=path, delete=True)
        result_file.close() # Close file to allow subprocess to write
        
        command = ['pytest', self.test_file, '--junitxml', result_file.name]
        cmd_result = run_command(command, cwd=path, timeout=self.timeout)
        cmd_result.check_returncode({ 0, 1 })

        report_xml = xml.etree.ElementTree.parse(result_file.name)

        return [part.grade(report_xml.find(f'.//testcase[@name=\'{part.name}\']')) for part in parts]



from fractions import Fraction
from typing import Literal
from typing_extensions import override

from ..utils import OptionalList, ShlexCommand, run_command
from ..grades import PartGrade
from . import GraderInterface, Part

"""
Grade a homework by execuing a bunch of commands
"""


class Command(Part):
    """A command to execute"""

    summary: str
    """
    A text description of the command (displayed on Gradescope)
    """

    command: ShlexCommand
    """
    The command to execute.
    """

    @override
    def description(self):
        return self.summary

    def grade(self, path, timeout):
        cmd_result = run_command(self.command, cwd=path, timeout=timeout)
        log = cmd_result.stdout

        if cmd_result.returncode:
            score = Fraction(0)
            log += f'\n\ngrader exited with exit code {cmd_result.returncode} != 0'
        else:
            score = Fraction(1)

        return PartGrade(score=score, log=log)


class MultiCommandGrader(GraderInterface[Command]):
    """
    Run a bunch of commands, testing the exit code of each.
    """

    kind: Literal["MultiCommandGrader"]
    
    timeout: float = 30
    """Timeout of autograder."""

    extra_setup_commands: OptionalList[ShlexCommand]

    @override
    def list_extra_setup_commands(self):
        return self.extra_setup_commands

    @override
    def Part(self, _cd):
        return Command

    def grade(self, submission, path, parts):
        return [part.grade(path, self.timeout) for part in parts]



from fractions import Fraction
from typing import Annotated

from pydantic import BaseModel, PlainSerializer, PlainValidator

from zucchini.exceptions import ZucchiniError

"""Store grades for components and parts."""

class PartGrade(BaseModel):
    """
    The result of grading a singular part.

    This is produced by Graders to indicate how complete a part is.
    """

    score: Fraction | int
    """The percentage correct (must be in [0, 1])."""

    deductions: list[str] | None = None
    """
    A list of strings which specify reasons points were deducted.
    This is purely descriptive and will be shown to the end-user.
    """

    log: str | None = None
    """Verbose logs after grading this part."""

class BoundPartGrade(BaseModel):
    """
    The result of grading a singular part and applying appropriate weights to it.
    """

    inner: PartGrade
    """
    Unweighted result.
    """

    description: str
    """
    Description of part (will be displayed on Gradescope)
    """

    norm_weight: Fraction
    """
    Normalized weight of part (within a component, is in [0, 1]).
    """

    def passed(self) -> bool:
        """Whether this part passed."""
        return self.inner.score == 1
    
    def points_received(self) -> Fraction:
        """The points received for this part, using normalized weight."""
        return self.inner.score * self.norm_weight

class ComponentGrade(BaseModel):
    """
    The result of grading an assignment component.
    """

    norm_weight: Fraction
    """
    Normalized weight of component (within an assignment, is in [0, 1]).
    """

    description: str | None = None
    """
    Description for component.
    """

    parts: list[BoundPartGrade] | None = None
    """
    Grades of each part in the component.

    This may be none if the component produced a `BrokenSubmissionError`.
    """

    error: Annotated[
        ZucchiniError,
        PlainValidator(ZucchiniError._validate),
        PlainSerializer(ZucchiniError._serialize)
    ] | None = None
    """
    Any submission errors which occurred during grading.

    This is `None` if no error occurred.
    """

    def passed(self) -> bool:
        """Whether this component passed."""
        return (
            self.error is not None
            and self.parts is not None
            and all(p.passed() for p in self.parts)
        )
    
    def points_received(self) -> Fraction:
        """The total number of points received for this component, using normalized weight and prior to any applied penalties."""
        if self.error or self.parts is None:
            return Fraction(0)
        return sum((p.points_received() for p in self.parts), start=Fraction(0)) * self.norm_weight
    
class PenaltyDeduction(BaseModel):
    name: str
    """Name of the deduction."""

    points_deducted: Fraction
    """Amount of points deducted due to this penalty, using normalized weight."""

class AssignmentGrade(BaseModel):
    """
    The result of grading the assignment.
    """

    name: str
    """
    The name of the assignment.
    """
    raw_score: Fraction
    """Normalized grade before penalties (in [0, 1])."""
    final_score: Fraction
    """Normalized grade after penalties (in [0, 1])."""

    max_points: Fraction
    """Maximum number of points."""

    components: list[ComponentGrade]
    """Grades from components."""

    penalties: list[PenaltyDeduction]
    """Any penalty deductions."""

    def final_grade(self) -> Fraction:
        """Final grade, scaled to the number of points."""
        return self.final_score * self.max_points



from ..exceptions import InvalidPenalizerConfigError
from .penalizer_interface import PenalizerInterface
from .late_penalizer import LatePenalizer

__all__ = ['InvalidPenalizerConfigError', 'PenalizerInterface',
           'LatePenalizer']

# If another penalizer is added, please update the format of below to match zucchini.graders.
_PENALIZERS = (
    LatePenalizer,
)
AVAILABLE_PENALIZERS = {cls.__name__: cls for cls in _PENALIZERS}



from pathlib import Path
from abc import abstractmethod
from typing import Generic
from typing_extensions import override
import concurrent.futures

from ..grades import PartGrade
from ..submission import Submission


from ..graders.grader_interface import P
from . import GraderInterface


"""
An implementation of GraderInterface which runs tests in separate
threads. In theory, speeds up grading by two times the number of
CPUs.

Actual graders can subclass this and implement grade_part() to become
threaded!
"""

class ThreadedGrader(GraderInterface[P], Generic[P]):
    """
    A base class for graders to run parts in separate threads to speed
    up grading. Subclasses must implement grade_part().
    """
    num_threads: int | None = None

    @abstractmethod
    def grade_part(self, part: P, path: Path, submission: Submission) -> PartGrade:
        """
        Grade a Part instance part, where the temporary grading
        directory is `path' and the Submission instance passed to
        grade() is `submission'.

        Must return a PartGrade instance.
        """
        pass

    @override
    def grade(self, submission: Submission, path: Path, parts: list[P]) -> list[PartGrade]:
        """Spin off the configured number of grading threads and grade."""

        grades: list[PartGrade | None] = [None] * len(parts)

        try:
            pool = concurrent.futures.ThreadPoolExecutor(thread_name_prefix="zgrader")
            # Create all futures
            futures = {pool.submit(self.grade_part, part, path, submission): i for i, part in enumerate(parts)}
            for fut in concurrent.futures.as_completed(futures):
                if (exc := fut.exception()) is not None:
                    raise exc
                
                # When a future is completed, insert it into grades
                index = futures[fut]
                grades[index] = fut.result()
        finally:
            pool.shutdown(wait=True, cancel_futures=True)

        return grades # type: ignore



from abc import ABC, abstractmethod
import functools
from typing import Callable, TextIO, TypeVar

from zucchini.grades import AssignmentGrade


class ExporterInterface(ABC):
    @abstractmethod
    def export(self, grade: AssignmentGrade, output: TextIO):
        """
        Take in a grade format and writes it to the output field.
        """
        pass

E = TypeVar("E", bound=ExporterInterface)
def _export_str(c: Callable[[E, AssignmentGrade], str]) -> Callable[[E, AssignmentGrade, TextIO], None]:
    """
    Decorator which takes a function that produces a string from the specified grade
    and converts it into the export format required by `ExporterInterface`.
    """

    @functools.wraps(c)
    def export(self: E, grade: AssignmentGrade, output: TextIO):
        output.write(c(self, grade))
    
    return export



class ZucchiniError(ValueError):
    """
    Any error which occurs when calling Zucchini.

    This could be misconfiguration on the autograder side or
    broken files on the user side.
    """
    def __init__(self, message: str, verbose: str | None = None, is_it_autograders_fault: bool = False):
        super().__init__(message)
        self.message = message
        self.verbose = verbose
        self.is_it_autograders_fault = is_it_autograders_fault

    @classmethod
    def _validate(cls, data: "dict | ZucchiniError"):
        if isinstance(data, ZucchiniError):
            return data
        
        error = data["error"]
        verbose = data["verbose"]
        autograder = data["autograder"]
        if autograder:
            return BrokenAutograderError(error, verbose)
        else:
            return BrokenSubmissionError(error, verbose)
        
    def _serialize(self):
        return {
            "error": self.message,
            "verbose": self.verbose,
            "autograder": self.is_it_autograders_fault
        }

class BrokenAutograderError(ZucchiniError):
    """
    Any error which occurs due to the autograder being misconfigured.
    """
    def __init__(self, message: str, verbose: str | None = None):
        super().__init__(message, verbose, is_it_autograders_fault=True)

class InvalidGraderConfigError(BrokenAutograderError):
    """
    Exception indicating the Zucchini config file
    had a misconfigured grader backend.
    """
    pass

class InvalidPenalizerConfigError(BrokenAutograderError):
    """
    Exception indicating the Zucchini config file
    had a misconfigured penalizer backend.
    """
    pass

class BrokenSubmissionError(ZucchiniError):
    """
    Exception indicating the submitted file could not be autograded
    (it is unable to be run or an error occurred while running).
    """
    def __init__(self, message: str, verbose: str | None = None):
        super().__init__(message, verbose, is_it_autograders_fault=False)


from typing import Literal
from typing_extensions import override

from zucchini.grades import AssignmentGrade
from zucchini.gradescope import GradescopeOutput

from .exporter_interface import ExporterInterface, _export_str

class LocalExporter(ExporterInterface):
    @override
    @_export_str
    def export(self, grade: AssignmentGrade) -> str:
        output_arr = []
        any_errors = False
        for component in grade.components:
            if component.error:
                any_errors = True
                output_arr.append(f"ERROR: {str(component.description):45} {component.error.message:15}")
                if component.error.verbose:
                    output_arr.append(f"Details: {component.error.verbose:45}")
            else:
                for part in (component.parts or []):
                    points_got = part.points_received() * component.norm_weight * grade.max_points
                    points_max = part.norm_weight * component.norm_weight * grade.max_points
                    points = f'{float(points_got):.2f}/{float(points_max):.2f}'
                    if part.passed():
                        output_arr.append(f"TEST: {part.description:45} {'PASSED':15} ({points})")
                    else:
                        output_arr.append(f"TEST: {part.description:45} {'FAILED':15} ({points})")
                        output_arr.append(part.inner.log)
        score = f'Total score: {float(100 * grade.final_score):.2f}%'
        output_arr.append(score)
        if any_errors:
            output_arr.append('Some errors occurred; the score above may not be'
                          ' your final grade')
        
        return "\n\n".join(output_arr)

class GradescopeExporter(ExporterInterface):
    @override
    @_export_str
    def export(self, grade: AssignmentGrade) -> str:
        return GradescopeOutput.from_grade(grade).model_dump_json()
        
ExporterKey = Literal["local", "gradescope"]
EXPORTERS: dict[ExporterKey, type[ExporterInterface]] = {
    "local": LocalExporter,
    "gradescope": GradescopeExporter
}


import dataclasses
import datetime as dt
from fractions import Fraction
from pathlib import Path
import tempfile
from typing import Annotated

from pydantic import BeforeValidator, Field, ValidationInfo

from zucchini.exceptions import BrokenSubmissionError, BrokenAutograderError, ZucchiniError
from zucchini.graders import SupportedGrader
from zucchini.graders.grader_interface import GraderInterface, Part
from zucchini.penalizers.late_penalizer import LatePenalizer
from zucchini.submission import Submission
from zucchini.utils import KebabModel, OptionalList, copy_globs

from .grades import AssignmentGrade, BoundPartGrade, ComponentGrade, PenaltyDeduction

def _div_or_zero(a: Fraction, b: int | Fraction):
    """
    Divides a by b, returning 0 if a division by zero error would occur.
    """
    if b == 0: return Fraction(0)
    return a / b

@dataclasses.dataclass(slots=True)
class AssignmentMetadata:
    total_points: Fraction = Fraction(100)
    """Total points the autograded portion of the assignment is worth."""

    due_date: dt.datetime | None = None
    """Date assignment is due."""

class Penalizer(KebabModel):
    """
    A penalizer config definition, which can use any defined `PenalizerInterface`.
    """

    name: str
    """Name of penalizer (displayed on Gradescope)"""

    # If we add more backends, switch this to:
    #     Annotated[SupportedPenalizer, Field(discriminator="kind")]
    backend: LatePenalizer
    """The backend to use"""

def _select_part(n, info: ValidationInfo):
    # Get backend:
    backend = info.data.get("backend")
    if not isinstance(backend, GraderInterface):
        raise ValueError("Could not resolve part due to misconfigured backend")

    return backend.create_part(n)

class AssignmentComponent(KebabModel):
    """
    A component which is graded with the same backend and files.
    """

    name: str
    """Name of component"""

    weight: Fraction
    """Weight of component"""

    backend: Annotated[SupportedGrader, Field(discriminator="kind")]
    """The backend to use"""
    
    files: OptionalList[str]
    """
    Paths (globs) copied from the submission folder.
    This should hold all files which are submitted from the user.
    """

    grading_files: OptionalList[str]
    """
    Paths (globs) copied from the grading folder.
    This should hold all files which are needed for grading.
    """

    parts: OptionalList[Annotated[Part, BeforeValidator(_select_part)]]

    def total_part_weight(self) -> Fraction:
        """Total weight defined by the parts of this component."""
        return sum((p.weight for p in self.parts), start=Fraction(0))

    def grade(self, submission: Submission, total_component_weight: Fraction, test_dir: Path) -> ComponentGrade:
        with tempfile.TemporaryDirectory(prefix="zcomponent-") as grading_dir:
            grading_dir = Path(grading_dir)
            parts_: list = self.parts
            norm_weight = _div_or_zero(self.weight, total_component_weight)

            try:
                try:
                    # Copy all submission files over
                    copy_globs(self.files, submission.submission_dir, grading_dir)
                except FileNotFoundError as e:
                    # This is user's fault
                    raise BrokenSubmissionError(str(e)) from e
            
                # Copy all grading files over
                copy_globs(self.grading_files, test_dir, grading_dir)

                # Perform grading:
                grades = self.backend.grade(submission, grading_dir, parts_)

            except ZucchiniError as e:
                return ComponentGrade(norm_weight=norm_weight, description=self.name, error=e)
            except Exception as e:
                # If something breaks, we assume autograder broke it:
                ze = BrokenAutograderError(str(e))
                return ComponentGrade(norm_weight=norm_weight, description=self.name, error=ze)
            
            total_part_weight = self.total_part_weight()
            parts = [
                BoundPartGrade(
                    inner=g,
                    description=p.description(),
                    norm_weight=_div_or_zero(p.weight, total_part_weight)
                )
                for (p, g) in zip(self.parts, grades)
            ]
            return ComponentGrade(norm_weight=norm_weight, description=self.name, parts=parts)


class AssignmentConfig(KebabModel):
    """
    A full assignment, which can be graded.
    """

    name: str
    """Name of assignment"""

    author: Annotated[
        OptionalList[str],
        # If string is provided (instead of a list), break up by commas:
        BeforeValidator(lambda s: [a.strip() for a in s.split(",")] if isinstance(s, str) else s)
    ]
    """Author of assignment"""

    penalties: OptionalList[Penalizer]
    """Any penalizers to impose on the assignment (e.g., late penalty)"""

    components: OptionalList[AssignmentComponent]
    """
    Individual components which use the same backend and file configuration.
    """

    def total_component_weight(self) -> Fraction:
        """Total weight defined by the components of this assignment."""
        return sum((c.weight for c in self.components), start=Fraction(0))


@dataclasses.dataclass
class Assignment:
    config: AssignmentConfig
    """
    The Zucchini configuration for the assignment,
    which specifies how it should be autograded.
    """

    grading_dir: Path
    """
    Directory where grading files are located.
    """

    metadata: AssignmentMetadata
    """
    The metadata for the assignment,
    which may be used to inform grading decisions.
    """
    
    def grade(self, submission: Submission):
        component_grades = [c.grade(submission, self.config.total_component_weight(), self.grading_dir) for c in self.config.components]
        penalties: list[PenaltyDeduction] = []
        
        raw_grade = sum((cg.points_received() for cg in component_grades), start=Fraction(0))
        adjusted_grade = raw_grade
        for p in self.config.penalties:
            new_adjusted_grade = p.backend.adjust_grade(adjusted_grade, submission, self.metadata)

            penalties.append(PenaltyDeduction(
                name=p.name,
                points_deducted=(adjusted_grade - new_adjusted_grade)
            ))
            adjusted_grade = new_adjusted_grade
        

        return AssignmentGrade(
            name=self.config.name,
            raw_score=raw_grade,
            final_score=adjusted_grade,
            max_points=self.metadata.total_points,
            components=component_grades,
            penalties=penalties
        )



import dataclasses
import itertools
from pathlib import Path
import shlex
from typing import Annotated, TypeAlias, TypeVar
import os
import shutil
import subprocess

from pydantic import BaseModel, BeforeValidator, ConfigDict, Field

from .exceptions import BrokenSubmissionError

@dataclasses.dataclass(init=False, slots=True)
class RunCommandResult:
    """
    A wrapper around `subprocess.CompletedProcess` to better support the needs of Zucchini graders.
    """

    returncode: int
    """The command's return code."""

    stdout: str
    """STDOUT, as a print-safe string"""

    stderr: str | None
    """
    STDERR, as a print-safe string (if it exists).
    STDERR will only exist if stderr was set to `subprocess.PIPE`.
    """

    def __init__(self, p: subprocess.CompletedProcess[bytes]):
        self.returncode = p.returncode
        self.stdout = p.stdout.decode(errors="backslashreplace")
        self.stderr = p.stderr.decode(errors="backslashreplace") if p.stderr is not None else None

    def check_returncode(self, acceptable_error_codes: set[int] = { 0 }):
        """
        Checks the return code, raising `BrokenSubmissionError` if not an acceptable error code.

        Parameters
        ----------
        acceptable_error_codes : set[int], optional
            The acceptable error codes (by default, { 0 })

        Raises
        ------
        BrokenSubmissionError
            When return code is not in the `acceptable_error_codes` set.
        """
        if self.returncode not in acceptable_error_codes:
            if self.stderr is None:
                verbose = self.stdout
            else:
                verbose = "\n".join([self.stdout, self.stderr])

            raise BrokenSubmissionError(
                f"grader command exited with exit code {self.returncode}.",
                verbose=verbose
            )

def run_command(
        cmd, *,
        cwd: Path | None = None,
        timeout: float | None = None,
        stderr = None,
        input = None,
    ):
    """
    Synchronously runs the command with `subprocess.run` and processes the output for grader use.

    Parameters
    ----------
    cmd
        The command, in `shlex.split` format
    cwd : Path
        The path of the current directory (cwd)
    timeout : float, optional
        The timeout before the command is canceled, by default None
    stderr
        Place to write STDERR. By default, it is piped into the STDOUT
        (which is itself piped as the output of this function).
    input
        Input to pass to the command (as STDIN).
    Returns
    -------
    RunCommandResult
        A result type which provides the result of the command execution.

    Raises
    ------
    BrokenSubmissionError
        If a timeout occurs.
    """
    try:
        p = subprocess.run(
            cmd,
            cwd=cwd,
            stdout=subprocess.PIPE,
            stderr=stderr or subprocess.STDOUT,
            timeout=timeout,
            input=input
        )
    except subprocess.TimeoutExpired:
        raise BrokenSubmissionError(f"grader timed out after {timeout} seconds")
    
    return RunCommandResult(p)

def _as_shlex_cmd(s: str | list[str]) -> list[str]:
    if isinstance(s, list):
        return s
    return shlex.split(s)
ShlexCommand: TypeAlias = Annotated[list[str], BeforeValidator(_as_shlex_cmd)]
"""
Pydantic validator which accepts strings (and lists of strings) which act as script commands,
and exposes the field as a split script command (as if from `shlex.split`).
"""

def sanitize_path(path: os.PathLike[str] | str) -> Path:
    """
    Convert an untrusted path to a relative path.
    """

    # Remove intermediate ..s
    normpath = Path(os.path.normpath(path))
    if normpath.is_absolute():
        # Remove absolute paths
        return Path(*normpath.parts[1:])
    else:
        # Remove leading ..s in relative paths
        parts = itertools.dropwhile(lambda p: p == os.path.pardir, normpath.parts)
        return Path(*parts)

def _copy_no_symlinks(src: os.PathLike[str] | str, dst: os.PathLike[str] | str):
    """
    Copies the source file to the destination (like `shutil.copy2`),
    but errors if source is a symlink.
    """
    if Path(src).is_symlink():
        raise ValueError(f"Cannot copy: {src} is a symlink")
    
    return shutil.copy2(src, dst)

def glob_get_files(globs: list[str], src_dir: Path) -> list[Path]:
    """
    Returns a list of paths obtained by the glob,
    returning an error if any of the glob paths don't match anything in the source directory.
    """
    files_to_copy: list[Path] = []

    # Do a first pass to check for missing files. This way, we don't
    # copy a bunch of files only to blow up when we can't find a
    # later file.
    for file_glob in globs:
        old_len = len(files_to_copy)
        files_to_copy += src_dir.glob(file_glob)

        if len(files_to_copy) - old_len == 0: # No new files were added
            raise FileNotFoundError(f"no file matches {file_glob!r} in {src_dir}")
    
    return files_to_copy

def copy_globs(globs: list[str], src_dir: os.PathLike[str], dest_dir: os.PathLike[str]):
    """
    Copy files matched by `globs` (a list of glob strings) from src_dir
    to dest_dir, maintaining directories if possible.
    """

    src_dir = Path(src_dir)
    dest_dir = Path(dest_dir)

    for src_file in glob_get_files(globs, src_dir):
        rel_path = src_file.relative_to(src_dir)
        dest = dest_dir / rel_path

        if src_file.is_dir():
            dest.mkdir(parents=True, exist_ok=True)
            shutil.copytree(src_file, dest, copy_function=_copy_no_symlinks, dirs_exist_ok=True)
        else:
            dest.parent.mkdir(parents=True, exist_ok=True)
            _copy_no_symlinks(src_file, dest)

T = TypeVar("T")
OptionalList = Annotated[list[T], Field(default_factory=list)]
"""
Pydantic type which allows for an optional list field.
If the specified field is missing, the field is filled with an empty list.
"""

class KebabModel(BaseModel):
    """
    A `pydantic.BaseModel`, 
    which converts snake case fields to kebab case (`_` => `-`) in serialization.
    """
    model_config = ConfigDict(alias_generator=lambda field: field.replace("_", "-"))
